'''
This module determine and expose various information about the java
environment.
'''

import sys
from os.path import join, exists, dirname, realpath
from os import getenv
from platform import machine
from subprocess import Popen, check_output, PIPE
from shlex import split


PY2 = sys.version_info.major < 3

machine = machine()  # not expected to change at runtime

# This dictionary converts values from platform.machine()
# to a "cpu" string. It is needed to set the correct lib path,
# found in the JRE_HOME, e.g.: <JRE_HOME>/lib/<cpu>/.
MACHINE2CPU = {
    "i686": "i386",
    "x86_64": "amd64",
    "AMD64": "amd64",
    "armv7l": "arm",
    "sun4u": "sparcv9",
    "sun4v": "sparcv9"
}

JAVA_HOME = getenv('JAVA_HOME')


def find_javac(platform, possible_homes):
    '''Find javac in all possible locations.'''
    name = "javac.exe" if platform == "win32" else "javac"
    for home in possible_homes:
        for javac in [join(home, name), join(home, 'bin', name)]:
            if exists(javac):
                if platform == "win32" and not PY2:  # Avoid path space execution error
                    return '"%s"' % javac
                return javac
    return name  # Fall back to "hope it's on the path"


def get_include_dirs(platform):
    if platform == 'darwin':
        framework = get_osx_framework()
        if '1.6' in framework:
            return [join(
                framework, (
                    'System/Library/Frameworks/'
                    'JavaVM.framework/Versions/Current/Headers'
                )
            )]
        else:
            # We want to favor Java installation declaring JAVA_HOME
            if JAVA_HOME:
                framework = JAVA_HOME

            return [
                '{0}/include'.format(framework),
                '{0}/include/darwin'.format(framework)
            ]

    else:
        jdk_home = get_jdk_home(platform)
        if platform == 'win32':
            incl_dir = join(jdk_home, 'include', 'win32')
        elif platform == 'sunos5':
            incl_dir = join(jdk_home, 'include', 'solaris')
        else:
            incl_dir = join(jdk_home, 'include', 'linux')

        return [
            join(jdk_home, 'include'),
            incl_dir
        ]

def get_library_dirs(platform, arch=None):
    if platform == 'win32':
        jre_home = get_jre_home(platform)
        jdk_home = JAVA_HOME

        if isinstance(jre_home, bytes):
            jre_home = jre_home.decode('utf-8')

        return [
            join(jdk_home, 'lib'),
            join(jdk_home, 'bin', 'server')
        ]
    elif platform == 'android':
        return ['libs/{}'.format(arch)]
    return []


def get_jre_home(platform):
    jre_home = None
    if JAVA_HOME and exists(join(JAVA_HOME, 'jre')):
        jre_home = join(JAVA_HOME, 'jre')

    if platform != 'win32' and not jre_home:
        jre_home = realpath(
            check_output(
                split('which java')
            ).decode('utf-8').strip()
        ).replace('bin/java', '')

    if platform == 'win32':
        if isinstance(jre_home, bytes):
            jre_home = jre_home.decode('utf-8')

    return jre_home


def get_jdk_home(platform):
    jdk_home = getenv('JDK_HOME')
    if not jdk_home:
        if platform == 'win32':
            TMP_JDK_HOME = getenv('JAVA_HOME')
            print(TMP_JDK_HOME)
            if not TMP_JDK_HOME:
                raise Exception('Unable to find JAVA_HOME')

            # Remove /bin if it's appended to JAVA_HOME
            if TMP_JDK_HOME[-3:] == 'bin':
                TMP_JDK_HOME = TMP_JDK_HOME[:-4]

            # Check whether it's JDK
            if exists(join(TMP_JDK_HOME, 'bin', 'javac.exe')):
                jdk_home = TMP_JDK_HOME

        else:
            jdk_home = realpath(
                check_output(
                    ['which', 'javac']
                ).decode('utf-8').strip()
            ).replace('bin/javac', '')

    if not jdk_home or not exists(jdk_home):
        raise Exception('Unable to determine JDK_HOME')

    return jdk_home


def get_osx_framework():
    framework = Popen(
        '/usr/libexec/java_home',
        stdout=PIPE, shell=True
    ).communicate()[0]

    if not PY2:
        framework = framework.decode('utf-8')

    return framework.strip()


def get_possible_homes(platform):
    if platform == 'darwin':
        if JAVA_HOME:
            return JAVA_HOME

        FRAMEWORK = get_osx_framework()
        if not FRAMEWORK:
            raise Exception('You must install Java for Mac OSX')

        return FRAMEWORK

    else:
        return (
            get_jdk_home(platform),
            get_jre_home(platform),
        )


def get_cpu():
    try:
        return MACHINE2CPU[machine]
    except KeyError:
        print(
            "WARNING: Not able to assign machine()"
            " = %s to a cpu value!" % machine
        )
        print("         Using cpu = 'i386' instead!")
        return 'i386'


def get_libraries(platform):
    if platform == 'android':
        # for android, we use SDL...
        return ['sdl', 'log']

    elif platform == 'win32':
        return ['jvm']


def get_jnius_lib_location(platform):
    cpu = get_cpu()

    if platform == 'darwin':
        framework = get_osx_framework()

        if '1.6' in framework:
            return '../Libraries/libjvm.dylib'

        else:
            lib_location = 'jre/lib/jli/libjli.dylib'

            # We want to favor Java installation declaring JAVA_HOME
            if JAVA_HOME:
                framework = JAVA_HOME

            full_lib_location = join(framework, lib_location)

            if not exists(full_lib_location):
                # In that case, the Java version is very likely >=9.
                # So we need to modify the `libjvm.so` path.
                lib_location = 'lib/jli/libjli.dylib'
                full_lib_location = join(framework, lib_location)

            if not exists(full_lib_location):
                # adoptopenjdk12 doesn't have the jli subfolder either
                return 'lib/libjli.dylib'

            return lib_location

    elif platform == 'sunos5':
        return 'jre/lib/{}/server/libjvm.so'.format(cpu)

    else:
        if platform not in ('linux', 'linux2'):
            print("warning: unknown platform assuming linux")

        lib_location = 'jre/lib/{}/server/libjvm.so'.format(cpu)

        jre_home = dirname(get_jre_home(platform)).strip('jre')

        full_lib_location = join(jre_home, lib_location)

        if not exists(full_lib_location):
            # In that case, the Java version is very likely >=9.
            # So we need to modify the `libjvm.so` path.
            lib_location = 'lib/server/libjvm.so'
        return lib_location



from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division
__all__ = ('autoclass', 'ensureclass')
from six import with_metaclass
import logging

from .jnius import (
    JavaClass, MetaJavaClass, JavaMethod, JavaStaticMethod,
    JavaField, JavaStaticField, JavaMultipleMethod, find_javaclass,
    JavaException
)

log = logging.getLogger(__name__)


class Class(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/Class'

    desiredAssertionStatus = JavaMethod('()Z')
    forName = JavaMultipleMethod([
        ('(Ljava/lang/String,Z,Ljava/lang/ClassLoader;)Ljava/langClass;', True, False),
        ('(Ljava/lang/String;)Ljava/lang/Class;', True, False), ])
    getClassLoader = JavaMethod('()Ljava/lang/ClassLoader;')
    getClasses = JavaMethod('()[Ljava/lang/Class;')
    getComponentType = JavaMethod('()Ljava/lang/Class;')
    getConstructor = JavaMethod('([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;')
    getConstructors = JavaMethod('()[Ljava/lang/reflect/Constructor;')
    getDeclaredClasses = JavaMethod('()[Ljava/lang/Class;')
    getDeclaredConstructor = JavaMethod('([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;')
    getDeclaredConstructors = JavaMethod('()[Ljava/lang/reflect/Constructor;')
    getDeclaredField = JavaMethod('(Ljava/lang/String;)Ljava/lang/reflect/Field;')
    getDeclaredFields = JavaMethod('()[Ljava/lang/reflect/Field;')
    getDeclaredMethod = JavaMethod('(Ljava/lang/String,[Ljava/lang/Class;)Ljava/lang/reflect/Method;')
    getDeclaredMethods = JavaMethod('()[Ljava/lang/reflect/Method;')
    getDeclaringClass = JavaMethod('()Ljava/lang/Class;')
    getField = JavaMethod('(Ljava/lang/String;)Ljava/lang/reflect/Field;')
    getFields = JavaMethod('()[Ljava/lang/reflect/Field;')
    getInterfaces = JavaMethod('()[Ljava/lang/Class;')
    getMethod = JavaMethod('(Ljava/lang/String,[Ljava/lang/Class;)Ljava/lang/reflect/Method;')
    getMethods = JavaMethod('()[Ljava/lang/reflect/Method;')
    getModifiers = JavaMethod('()[I')
    getName = JavaMethod('()Ljava/lang/String;')
    getPackage = JavaMethod('()Ljava/lang/Package;')
    getProtectionDomain = JavaMethod('()Ljava/security/ProtectionDomain;')
    getResource = JavaMethod('(Ljava/lang/String;)Ljava/net/URL;')
    getResourceAsStream = JavaMethod('(Ljava/lang/String;)Ljava/io/InputStream;')
    getSigners = JavaMethod('()[Ljava/lang/Object;')
    getSuperclass = JavaMethod('()Ljava/lang/Class;')
    isArray = JavaMethod('()Z')
    isAssignableFrom = JavaMethod('(Ljava/lang/reflect/Class;)Z')
    isInstance = JavaMethod('(Ljava/lang/Object;)Z')
    isInterface = JavaMethod('()Z')
    isPrimitive = JavaMethod('()Z')
    newInstance = JavaMethod('()Ljava/lang/Object;')
    toString = JavaMethod('()Ljava/lang/String;')

    def __str__(self):
        return (
            '%s: [%s]' if self.isArray() else '%s: %s'
        ) % (
            'Interface' if self.isInterface() else
            'Primitive' if self.isPrimitive() else
            'Class',
            self.getName()
        )

    def __repr__(self):
        return '<%s at 0x%x>' % (self, id(self))


class Object(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/Object'

    getClass = JavaMethod('()Ljava/lang/Class;')
    hashCode = JavaMethod('()I')


class Modifier(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/reflect/Modifier'

    isAbstract = JavaStaticMethod('(I)Z')
    isFinal = JavaStaticMethod('(I)Z')
    isInterface = JavaStaticMethod('(I)Z')
    isNative = JavaStaticMethod('(I)Z')
    isPrivate = JavaStaticMethod('(I)Z')
    isProtected = JavaStaticMethod('(I)Z')
    isPublic = JavaStaticMethod('(I)Z')
    isStatic = JavaStaticMethod('(I)Z')
    isStrict = JavaStaticMethod('(I)Z')
    isSynchronized = JavaStaticMethod('(I)Z')
    isTransient = JavaStaticMethod('(I)Z')
    isVolatile = JavaStaticMethod('(I)Z')


class Method(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/reflect/Method'

    getName = JavaMethod('()Ljava/lang/String;')
    toString = JavaMethod('()Ljava/lang/String;')
    getParameterTypes = JavaMethod('()[Ljava/lang/Class;')
    getReturnType = JavaMethod('()Ljava/lang/Class;')
    getModifiers = JavaMethod('()I')
    isVarArgs = JavaMethod('()Z')


class Field(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/reflect/Field'

    getName = JavaMethod('()Ljava/lang/String;')
    toString = JavaMethod('()Ljava/lang/String;')
    getType = JavaMethod('()Ljava/lang/Class;')
    getModifiers = JavaMethod('()I')


class Constructor(with_metaclass(MetaJavaClass, JavaClass)):
    __javaclass__ = 'java/lang/reflect/Constructor'

    toString = JavaMethod('()Ljava/lang/String;')
    getParameterTypes = JavaMethod('()[Ljava/lang/Class;')
    getModifiers = JavaMethod('()I')
    isVarArgs = JavaMethod('()Z')


def get_signature(cls_tp):
    tp = cls_tp.getName()
    if tp[0] == '[':
        return tp.replace('.', '/')
    signatures = {
        'void': 'V', 'boolean': 'Z', 'byte': 'B',
        'char': 'C', 'short': 'S', 'int': 'I',
        'long': 'J', 'float': 'F', 'double': 'D'}
    ret = signatures.get(tp)
    if ret:
        return ret
    # don't do it in recursive way for the moment,
    # error on the JNI/android: JNI ERROR (app bug): local reference table
    # overflow (max=512)

    # ensureclass(tp)
    return 'L{0};'.format(tp.replace('.', '/'))


registers = []


def ensureclass(clsname):
    if clsname in registers:
        return
    jniname = clsname.replace('.', '/')
    if MetaJavaClass.get_javaclass(jniname):
        return
    registers.append(clsname)
    autoclass(clsname)


def lower_name(s):
    return s[:1].lower() + s[1:] if s else ''


def bean_getter(s):
    return (s.startswith('get') and len(s) > 3 and s[3].isupper()) or (s.startswith('is') and len(s) > 2 and s[2].isupper())


def log_method(method, name, signature):
    mods = method.getModifiers()
    log.debug(
        '\nmeth: %s\n'
        '  sig: %s\n'
        '  Public %s\n'
        '  Private %s\n'
        '  Protected %s\n'
        '  Static %s\n'
        '  Final %s\n'
        '  Synchronized %s\n'
        '  Volatile %s\n'
        '  Transient %s\n'
        '  Native %s\n'
        '  Interface %s\n'
        '  Abstract %s\n'
        '  Strict %s\n',
        name,
        signature,
        Modifier.isPublic(mods),
        Modifier.isPrivate(mods),
        Modifier.isProtected(mods),
        Modifier.isStatic(mods),
        Modifier.isFinal(mods),
        Modifier.isSynchronized(mods),
        Modifier.isVolatile(mods),
        Modifier.isTransient(mods),
        Modifier.isNative(mods),
        Modifier.isInterface(mods),
        Modifier.isAbstract(mods),
        Modifier.isStrict(mods)
    )


def autoclass(clsname):
    jniname = clsname.replace('.', '/')
    cls = MetaJavaClass.get_javaclass(jniname)
    if cls:
        return cls

    classDict = {}

    # c = Class.forName(clsname)
    c = find_javaclass(clsname)
    if c is None:
        raise Exception('Java class {0} not found'.format(c))
        return None

    constructors = []
    for constructor in c.getConstructors():
        sig = '({0})V'.format(
            ''.join([get_signature(x) for x in constructor.getParameterTypes()]))
        constructors.append((sig, constructor.isVarArgs()))
    classDict['__javaconstructor__'] = constructors

    cls = c
    while cls is not None:
        methods = cls.getDeclaredMethods()
        methods_name = [x.getName() for x in methods]

        for index, method in enumerate(methods):
            name = methods_name[index]
            if name in classDict:
                continue

            # only one method available
            if methods_name.count(name) == 1:
                static = Modifier.isStatic(method.getModifiers())
                varargs = method.isVarArgs()
                sig = '({0}){1}'.format(
                    ''.join([get_signature(x) for x in method.getParameterTypes()]),
                    get_signature(method.getReturnType()))
                if log.level <= logging.DEBUG:
                    log_method(method, name, sig)
                classDict[name] = (JavaStaticMethod if static else JavaMethod)(sig, varargs=varargs)
                if name != 'getClass' and bean_getter(name) and len(method.getParameterTypes()) == 0:
                    lowername = lower_name(name[2 if name.startswith('is') else 3:])
                    classDict[lowername] = (lambda n: property(lambda self: getattr(self, n)()))(name)
                continue

            # multiple signatures
            signatures = []
            for index, subname in enumerate(methods_name):
                if subname != name:
                    continue
                method = methods[index]
                sig = '({0}){1}'.format(
                    ''.join([get_signature(x) for x in method.getParameterTypes()]),
                    get_signature(method.getReturnType()))

                if log.level <= logging.DEBUG:
                    log_method(method, name, sig)
                signatures.append((sig, Modifier.isStatic(method.getModifiers()), method.isVarArgs()))

            classDict[name] = JavaMultipleMethod(signatures)

        _cls = cls.getSuperclass()
        if not _cls and cls.isInterface():
            cls = find_javaclass('java.lang.Object')
        else:
            cls = _cls

    def _getitem(self, index):
        try:
            return self.get(index)
        except JavaException as e:
            # initialize the subclass before getting the Class.forName
            # otherwise isInstance does not know of the subclass
            mock_exception_object = autoclass(e.classname)()
            if find_javaclass("java.lang.IndexOutOfBoundsException").isInstance(mock_exception_object):
                # python for...in iteration checks for end of list by waiting for IndexError
                raise IndexError()
            else:
                raise

    for iclass in c.getInterfaces():
        if iclass.getName() == 'java.util.List':
            classDict['__getitem__'] = _getitem
            classDict['__len__'] = lambda self: self.size()
            break

    for field in c.getFields():
        static = Modifier.isStatic(field.getModifiers())
        sig = get_signature(field.getType())
        cls = JavaStaticField if static else JavaField
        classDict[field.getName()] = cls(sig)

    classDict['__javaclass__'] = clsname.replace('.', '/')

    return MetaJavaClass.__new__(
        MetaJavaClass,
        clsname,  # .replace('.', '_'),
        (JavaClass, ),
        classDict)
